/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@weaver.org> except where otherwise
   indicated.
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This contains my homebrew FTP client implementation
 * It's crap, probably work with about 2 different servers, is full of 
 * buffer over-runs everywhere, and is crap */
 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "sitecopy.h"
#include "ftp.h"

extern enum debug_types debug_level;

char *remote_root;
char *local_root;

char *ftp_error;

int pi_socket; /* The socket for the PI (control) connection */
int dtp_socket; /* The socket for the DTP (data) connection */

/* The address of the DTP server - this goes here because I get
 * it from within an ftp_read() rather than in a decent manner */
struct sockaddr_in dtp_host; 

int ftp_close( ) {
    return ftp_exec( "QUIT" );
}

/* Changes to given directory
 * Gives appropriate FTP state response */
int ftp_chdir( char *dir ) {
    char command[BUFSIZ];
    sprintf( command, "CWD %s", dir );
    return ftp_exec( command );
}

/* Creates the given directory
 * FTP state response */
int ftp_mkdir( char *dir ) {
    char command[BUFSIZ];
    sprintf( command, "MKD %s%s", remote_root, dir );
    return ftp_exec( command );
}

/* Changes to given directory, creating if necessary and possible */
int ftp_godir( char *dir ) {
    int ret;
    ret = ftp_chdir( dir ) != FTP_OK;
    if( ret != FTP_OK ) {
	/* Attempt to create... */
	ret = ftp_mkdir( dir );
	if( ret == FTP_OK )
	    /* Attempt to chdir... */
	    ret = ftp_chdir( dir );
    }
    return ret;
}

/* Establishes the control connection, to hostname on portnum */
int ftp_connect( char *hostname, int portnum ) {
    
    struct sockaddr_in sa;
    struct hostent *hp;
    
    printf( "Connecting to host %s... ", hostname );
    
    if( ( hp = gethostbyname(hostname) ) == NULL ) {
	printf("\nUnknown host; %s.\n", hostname);
	exit(1);
    }
    
    if( ( pi_socket = socket(AF_INET, SOCK_STREAM, 0) ) < 0) {
	perror("socket");
	return FTP_ERROR;
    }
    
    sa.sin_family = AF_INET;
    sa.sin_port = htons(portnum); /* ftp port, host -> net byte orders */
    memcpy( &sa.sin_addr, hp->h_addr, hp->h_length );
    
    if( connect(pi_socket, (struct sockaddr *) &sa, sizeof(struct sockaddr_in)) < 0 ) {
	perror("connect");
	return FTP_ERROR;
    }
    
    printf( "connected.\n" );
    return ftp_read();
}

/* upload the given file */
int ftp_put( char *localfile, char *remotefile ) {
    int ret;
    char command[BUFSIZ];
    strcpy( command, "TYPE I" );
    ftp_exec( command );
    /* Go into passive mode */
    ret = ftp_exec( "PASV" );
    if( ret == FTP_PASSIVE ) {
	/* Yo! We got passive mode */
	if( ftp_connect_pasv() ) {
	    sprintf( command, "STOR %s%s", remote_root, remotefile );
	    ret = ftp_exec( command );
	    if( ret == FTP_READY ) {
		/* send the file */
		strcpy( command, local_root );
		strcat( command, localfile );
		send_file( command );
		close( dtp_socket );
		/* sent okay? */
		ret = ftp_read();
	    } else {
		fprintf( stderr, "Could not send file.\n" );
		close( dtp_socket );
	    }
	} else {
	    fprintf( stderr, "Could not establish data connection.\n");
	}
    } else {
	/* Swine, give up without PASV mode. */
	fprintf( stderr, "Error: Server does not support passive mode.\n");
    }
    return ret;
}

int ftp_delete( char *filename ) {
    char command[BUFSIZ];
    sprintf( command, "DELE %s%s", remote_root, filename );
    return ftp_exec( command );
}

int ftp_rmdir( char *filename ) {
    char command[BUFSIZ];
    sprintf( command, "RMD %s%s", remote_root, filename );
    return ftp_exec( command );
}

/* Send the fiven filename down the data connection 
 * Returns non-zero value if successful */
int send_file( char *filename ) {
    char buffer[BUFSIZ];
    int len, wrlen, fp;
    if( (fp = open( filename, O_RDONLY )) == -1 ) {
	perror("open"); 
	return 0;
    }
    printf("Sending %s ", filename);

    /* dump from file to socket */
    do {
	len = read( fp, buffer, BUFSIZ );
	if( len < 0 ) perror("read");
	if( len > 0 ) wrlen = write( dtp_socket, buffer, len );
	putc( '.', stdout );
    } while( len > 0 );

    close( fp );
    printf(" sent.\n");
    return FTP_OK;
}

/* Connect to the given host 
 * Returns non-zero value if successfull */
int ftp_connect_pasv() {
    if ( (dtp_socket = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
	perror("socket");
	return 0;
    }
    if( connect( dtp_socket , (struct sockaddr *) &dtp_host, sizeof(struct sockaddr_in)) < 0 ) {
	perror("connect");
	return 0;
    }
    return 1;
}

/* Do a login */
int ftp_login( char *username, char *password ) {
    char cmd[BUFSIZ];
    int ret;
    if( debug_level > debug_none )
	fprintf( stderr, "Logging in as %s...\n", username);
    sprintf( cmd, "USER %s", username );
    ret = ftp_exec( cmd );
    if( ret == FTP_NEEDPASSWORD ) {
	if( debug_level > debug_none )
	    fprintf( stderr, "Supplying password...\n" );
	sprintf( cmd, "PASS %s", password );
	ret = ftp_exec( cmd );
    } else {
	if( debug_level > debug_none )
	    fprintf( stderr, "Password not required.\n");
    }
    return ret;
}

int ftp_exec( char *command ) {
    char *buffer;
    /* This keeps causing me hassle, think it's right now. +1 for the \0 */
    buffer = malloc( strlen(command) + strlen("\r\n") + 1 );
    strcpy( buffer, command );
    /* Add \r\n on the end - Unix FTP servers will ignore the \r, 
     * Windows ones require \r\n */
    strcat( buffer, "\r\n" );
    if( debug_level > debug_none) printf(">%s\n", command );
    send_line( pi_socket, buffer );
    free( buffer );
    return ftp_read();
}

int ftp_read() {
    
    int multiline, len, reply_code;
    char buffer[BUFSIZ];

    multiline = 0;

    for(;;) {
	if( (len = read_line(pi_socket, buffer, BUFSIZ )) < 0 ) {
	    fprintf( stderr, "Could not read from socket." );
	    return FTP_ERROR;
	}
	
	if( debug_level > debug_none) printf("<%s",buffer);
	if(len<5) /* this is a multi-liner, ignore it and carry on */
	    continue; 
	/* parse the reply code from the line */
	reply_code = get_reply_code(buffer);
	
	/* If we have a VALID reply code and we are currently
	 * in a multi line response then this is the end of the
	 * multi line response */
	if(multiline && reply_code)
	    multiline=0;
	
	/* Now, if we aren't in a multi line response... */
	if(!multiline) { 
	    if(buffer[3]=='-') {
		/* A dash in char four denotes beginning of multi
		 * line response  'xxx-' */
		multiline=1;
	    } else {
		/* Looks like we've got a real response line */
		return ftp_response( buffer, reply_code );
	    }
	}
    }
    return FTP_ERROR;
}

int ftp_response( char *response, int code ) {
    ftp_error = response+4;
    switch( code ) {
    case 200: /* misc OK codes */
    case 220:
    case 230:
    case 250: /* completed file action */
    case 226: /* received file okay */
	return FTP_OK;
    case 150: /* ready to recv data */
	return FTP_READY;
    case 550: /* couldn't complete file action */
	return FTP_FILEBAD;
    case 331: /* got username, want password */
	return FTP_NEEDPASSWORD;
    case 221:
	return FTP_CLOSED;
    case 421: /* service denied */
	return FTP_DENIED;
    case 227: /* Passive mode */
	if( ftp_read_pasv( response ) )
	    return FTP_PASSIVE;
	else
	    return FTP_ERROR;
    default:
	return FTP_ERROR;
    }
}

int ftp_read_pasv( char *response ) {
    int h1, h2, h3, h4, p1, p2;
    char *start;
    start = strchr( response, '(' );
    /* get the host + port */
    if( sscanf( ++start, "%d,%d,%d,%d,%d,%d", &h1, &h2, &h3, &h4, &p1, &p2 ) < 6 )
	/* didn't match, give up */
	return 0;
    /* fill in the gory details */
    dtp_host.sin_family = AF_INET;
    /* not forgetting to convert host -> network byte orders... */
    dtp_host.sin_addr.s_addr = htonl( (h1<<24) | (h2<<16) | (h3<<8) | h4 );
    dtp_host.sin_port = htons( (p1<<8) | p2 );
    return 1;
}

void send_line(int sock, char *buffer) {
    write(sock, buffer, strlen(buffer));
}

int get_reply_code(char *buffer) {
    if(strlen(buffer) > 3)
	if( isdigit(buffer[0]) && isdigit(buffer[1]) && isdigit(buffer[2]) )
	    /* looks good */
	    return atoi(buffer);
    return 0;
}

/* This is from from Eric Raymonds fetchmail (SockRead() in socket.c)
 * since I wouldn't have a clue how to do it properly.
 * This function is Copyright (C) Eric Raymond.
 */
int read_line(int sock, char *buffer, int len) {
    char *newline, *bp = buffer;
    int n;
    
    if (--len < 1)
	return(-1);
    do {
	/*
	 * The reason for these gymnastics is that we want two things:
	 * (1) to read \n-terminated lines,
	 * (2) to return the true length of data read, even if the
	 *     data coming in has embedded NULS.
	 */
	if ((n = recv(sock, bp, len, MSG_PEEK)) <= 0)
	    return(-1);
	if ((newline = memchr(bp, '\n', n)) != NULL)
	    n = newline - bp + 1;
	if ((n = read(sock, bp, n)) == -1)
	    return(-1);
	bp += n;
	len -= n;
    } while (!newline && len);
    *bp = '\0';
    return bp - buffer;
}
