/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@weaver.org>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include "sitecopy.h"
#include "ftp.h"
#include "sites.h"

#define RCNAME "/.sitecopyrc"
#define COPYNAME "/.sitecopy/"

enum debug_types debug_level;
enum action_types action;

/* From the FTP module */
extern char *remote_root;
extern char *local_root;

/* For the site storage */
struct ll_sites *all_sites;
struct site_t *the_site;
char *sitename;

/* Misc global vars */
char *progname; /* argv[0] */
char *rcfile; /* filename of sitecopyrc */
char *copypath; /* path to sitecopy directory */
int confirms; /* Do we want confirmations? */

void init( int argc, char *argv[] ) {
    char *home;
    int optc;
    extern char *optarg;
    extern int optind, opterr, optopt;
    /* Defaults */
    action = action_list;
    debug_level = debug_none;
    progname = argv[0];
    while( (optc = getopt( argc, argv, "d:hrucil" ) ) != -1) {
	switch( optc ) {
	case 'l': action = action_list; break;
	case 'r': action = action_resync; break;   
	case 'i': action = action_init; break;
	case 'c': action = action_catchup; break;
	case 'u': action = action_update; break;
	case 'd':
	    switch( atoi(optarg) ) {
	    case 0: 
		debug_level = debug_none;
		printf( "Debugging off.\n" );
		break;
	    case 1:
		debug_level = debug_some;
		printf( "Debugging some.\n" );
		break;
	    case 2: 
		debug_level = debug_full; 
		printf( "Debugging full.\n" );
		break;
	    default:
		usage( ); exit(1);
	    }
	    break;
	case 'h': usage( ); exit(1);
	case '?': usage( ); exit(1);
	}
    }
    if( (argc - optind) != 1) {
	usage( argv[0] );
	exit(1);
    }
    sitename = argv[optind];
    /* Now get the sitecopyrc filename */
    home = getenv("HOME");
    /* Add 1 for the \0 */
    rcfile = (char *) malloc( strlen(RCNAME) + strlen(home) + 1);
    copypath = (char *) malloc( strlen(COPYNAME) + strlen(home) + 1);
    strcpy( rcfile, home );
    strcat( rcfile, RCNAME );
    strcpy( copypath, home );
    strcat( copypath, COPYNAME );
}

int main( int argc, char *argv[] ) {
    /* Initialize up */
    init( argc, argv );
    /* Read the rcfile */
    if( site_readsites( rcfile, &all_sites) ) {
	fprintf( stderr, "%s: Error: Corrupt rcfile.\n", progname );
	exit(1);
    }

    /* Find the site they want */
    if ( site_find( &all_sites, &the_site, sitename) ) {
	fprintf( stderr, "%s: Error: No site called `%s' found in rc file (%s)\n", progname, sitename, rcfile );
	exit(1);
    }
    printf( "%s: Site `%s' (%s:%s) in %s\n", progname, sitename, the_site->server, the_site->remote_dir, the_site->local_dir );
    remote_root = the_site->remote_dir;
    local_root = the_site->local_dir;
    if( site_readfiles( the_site ) ) {
	fprintf( stderr, "%s: Error: Could not read site files.\n", progname );
	exit(1);
    }
    /* Connect to the server */
    switch( action ) {
    case action_update:	site_update( the_site ); break;
    case action_list: site_list( the_site );break;
/*    case action_resync:	site_resync( the_site ); break; */
    case action_init:
	printf("Site initialized.\n");
	the_site->files = NULL; /* `Bit' of a hack */
	site_writefiles( the_site );
	break;
    case action_catchup:
	printf( "Catching up... " );
	site_writefiles( the_site );
	printf( "done.\nAll files and directories are marked as updated remotely\n" );
	break;
    default:
	printf("Unimplemented - sorry.\n"); break;
    }
    /* Clean up a bit */
    free( rcfile );
    free( copypath );
    return 0;
}

void usage( ) {
    fprintf( stderr,
"Usage: %s [-d LEVEL] [-h] [-l|-i|-u|-c] SITENAME\n\
\n\
\t-h\tDisplay this help message\n\
\t-d\tSet debugging level\n\
\t-l\tList files which have been added, changed, and deleted\n\
\t-i\tInitialize - mark all files and directories as not updated\n\
\t-c\tCatch up - mark all files and directories as updated\n\
\t-u\tUpdate the remote site\n\n\
This is sitecopy %s\n\n", progname, VERSION );

/* \t-r\tResyncronize the remote site list\n\
   \t-n\tDo no ask for confirmations\n\ */

}
