/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@weaver.org>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>

#include "sites.h"
#include "ftp.h"
#include "sitecopy.h"

#define DIRWORD "dir"

extern enum debug_types debug_level;
extern char *progname;
extern char *copypath;

extern char *ftp_error;

/* Temporary for use in ftw callback */
struct site_t *cur_site;

/* Functions for dealing with sites */

/* site_readsites will read ~/.sitecopyrc and fill the sites list */
int site_readsites( char *filename, struct ll_sites **sites ) {
    FILE *fp; int state, alpha;
    char buf[BUFSIZ];
    char *ch, *ptr, *key, *val;
    /* Holders for the site info */
    struct ll_sites *this_site, *last_site;

    if( (fp = fopen( filename, "r" )) == NULL ) {
	perror("fopen");
	return 1;
    } 

    key = malloc( BUFSIZ );
    val = malloc( BUFSIZ );

    last_site = this_site = NULL;

    while( fgets( buf, sizeof(buf), fp ) != NULL ) {
	state = 0;
	ptr = key;
	*key = *val = '\0';
	for( ch=buf; *ch!='\0'; ch++ ) {
	    alpha = !isspace(*ch);
	    switch( state ) {
	    case 0: 
		if(alpha) {
		    *(ptr++) = *ch;
		    state = 1;
		}
		break;
	    case 1:
		if(!alpha) {
		    *ptr = '\0';
		    ptr = val;
		    state = 2;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 2:
		if(alpha) {
		    *(ptr++) = *ch;
		    state = 3;
		}
		break;
	    case 3:
		if(!alpha) {
		    *ptr = '\0';
		    state = 4;
		} else {
		    *(ptr++) = *ch;
		}
	    default: /* ignores everything past end of value */
		break;
	    }
	}
	/* printf("Key [%s] Value: [%s]\n", key, val); */
	if( strlen( key ) > 0 ) {
	    if( strcmp( key, "site" ) == 0 ) {
		/* Beginning of a new Site */
		last_site = this_site;
		/* Allocate new item */
		this_site = malloc( sizeof( struct ll_sites ) );
		/* Zero it out */
		memset( this_site, 0, sizeof( struct ll_sites ) );
		if( last_site != NULL ) { /* next site */
		    last_site->next = this_site;
		} else { /* First site */
		    *sites = this_site;
		}
		this_site->item.name = strdup( val );
		this_site->item.files = NULL;
		/* Now work out the info filename */
		this_site->item.infofile = malloc( strlen(copypath) + strlen(val) + 1);
		strcpy( this_site->item.infofile, copypath );
		strcat( this_site->item.infofile, val );
	    } else if( this_site == NULL ) {
		/* TODO: Only do this if we have a valid keyword */
		fprintf( stderr, "Error: %s before site keyword\n", key );
	    } else if( strcmp( key, "username" ) == 0 ) {
		/* username */
		this_site->item.username = strdup( val );
	    } else if( strcmp( key, "server" ) == 0 ) {
		this_site->item.server = strdup( val );
	    } else if( strcmp( key, "password" ) == 0 ) {
		this_site->item.password = strdup( val );
	    } else if( strcmp( key, "remote" ) == 0 ) {
		if( val[strlen(val)-1] == '/' )
		    val[strlen(val)-1] = '\0';
		this_site->item.remote_dir = strdup( val );
	    } else if( strcmp( key, "local" ) == 0 ) {
		if( val[strlen(val)-1] == '/' )
		    val[strlen(val)-1] = '\0';
		this_site->item.local_dir = strdup( val );
	    } else {
		fprintf( stderr, "Error: Unknown keyword %s!\n", key );
		return 1;
	    }
	}
    }
    free(key);
    free(val);
    fclose( fp );
    return 0;
}

int site_find( struct ll_sites **sites, struct site_t **the_site, char *sitename ) {
    struct ll_sites *ptr;
    if( sites == NULL) return 1; /* be sane */
    ptr = *sites;
    while( ptr != NULL ) {
	/* printf("(searching: %s)\n", ptr->item.name ); */
	if( strcmp( ptr->item.name, sitename ) == 0 ) {
	    /* We found it */
	    *the_site = &(ptr->item);
	    return site_verify( *the_site );
	}
	ptr = ptr->next;
    }
    return 1;
}

int site_verify( struct site_t *any_site ) {
    if( any_site->name == NULL ) {
	fprintf( stderr, "Site name not specified.\n" );
    } else if( any_site->server == NULL ) {
	fprintf( stderr, "%s: Server not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->username == NULL ) {
	fprintf( stderr, "%s: Username not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->password == NULL ) {
	fprintf( stderr, "%s: Password not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->remote_dir == NULL ) {
	fprintf( stderr, "%s: Remote directory not specified in site `%s'.\n", progname, any_site->name );
    } else if( any_site->local_dir == NULL ) {
	fprintf( stderr, "%s: Local directory not specified in site `%s'.\n", progname, any_site->name );
    } else {
	return 0;
    }
    return 1;
}

/* Updates the remote files list of the given site.
 * Tricky, this - have to do a load of LIST's, and try and
 * parse the response sensibly */
int site_resync( struct site_t *the_site )  {
    site_writefiles( the_site );
    return 0;
}

int site_list( struct site_t *the_site ) {
    struct ll_files *current;
    int count;
    printf( "* The following files have been added since the last update:\n" );
    count = 0;
    for( current = the_site->files; current!=NULL; current=current->next ) {
	if( current->item.diff == file_new ) {
	    if( count++ ) printf( ", " );
	    if( current->item.dir ) printf( "dir:" );
	    printf( "%s", current->item.name);
	}
    }
    if( !count ) printf( "(none)" );
    printf("\n");
    printf( "* The following files have been changed since the last update:\n" );
    count = 0;
    for( current = the_site->files; current!=NULL; current=current->next ) {
	if( current->item.diff == file_changed ) {
	    if( count++ ) printf( ", " );
	    if( current->item.dir ) printf( "dir:" );
	    printf( "%s", current->item.name);
	}
    }
    if( !count ) printf( "(none)" );
    printf("\n");
    printf( "* These files have been deleted since the last update:\n" );
    count = 0;
    for( current = the_site->files; current!=NULL; current=current->next ) {
	if( current->item.diff == file_deleted ) {
	    if( count++ ) printf( ", " );
	    if( current->item.dir ) printf( "dir:" );
	    printf( "%s", current->item.name);
	}
    }
    if( !count ) printf( "(none)" );
    printf("\n");

    return 0;
}

/* Updates the remote site with the local copy.
 * This is the core functionality of sitecopy.
 * All we do is compare the local files list with the remote files
 * list, and upload any changed or new files. */
int site_update( struct site_t *the_site )  {
    int ret;
    struct ll_files *current;

    ret = ftp_connect( the_site->server, 21 );
    if( ret == FTP_OK ) {
	printf("Logging in... ");
	ret = ftp_login( the_site->username, the_site->password );
    } else {
	fprintf( stderr, "Could not connect to server.\n" );
    }
    if( ret == FTP_OK ) {
	/* Just do it */
	printf( "done.\n" );
	for( current=the_site->files; current!=NULL; current=current->next ) {
	    if( current->item.dir && (current->item.diff == file_new ) ) {
		/* New dir! */
		printf( "Creating directory: %s\n", current->item.name );
		ftp_mkdir( current->item.name );
	    }
	}
	for( current=the_site->files; current!=NULL; current=current->next ) {
	    if( !current->item.dir ) {
		switch( current->item.diff ) {
		case file_new: /* File is new, upload it */
		case file_changed: /* File has changed, upload it */
		    printf( "Uploading file: %s\n", current->item.name );
		    if( ftp_put( current->item.name, current->item.name ) != FTP_OK )
			printf( "Could not upload file: %s\n", ftp_error );
		    break;
		case file_deleted: /* File has been removed, delete it */
		    printf( "Deleting file: %s\n", current->item.name );
		    if( ftp_delete( current->item.name ) != FTP_OK ) 
			printf( "Could not delete file: %s\n", ftp_error );
		    break;
		default: /* Ignore everything else */
		    break;
		}
	    }
	}
	for( current=the_site->files; current!=NULL; current=current->next ) {
	    if( current->item.dir && (current->item.diff == file_deleted ) ) {
		printf( "Deleting directory: %s\n", current->item.name );
	        ftp_rmdir( current->item.name );
	    }

	}
    } else {
	fprintf( stderr, "Could not log in.\n" );
    }
 
    /* Close the connection */
    printf( "Closing connection... " );
    ftp_close();
    printf( "done.\n" );
    site_writefiles( the_site );
    return 0;
}

/* This reads off the remote files and the local files */
int site_readfiles( struct site_t *the_site ) {
    FILE *fp;
    char buf[BUFSIZ], tmp[BUFSIZ];
    struct ll_files *current;
    fp = fopen( the_site->infofile, "r" );
    if( fp != NULL ) {
	/* The file exists, so read it... */
	while( fgets( buf, BUFSIZ, fp) != NULL ) {
	    current = malloc( sizeof(struct ll_files) );
	    current->next = the_site->files;
	    the_site->files = current;
	    current->item.name = malloc( BUFSIZ );
	    sscanf( buf, "%s %s", current->item.name, tmp );
	    current->item.diff = file_deleted;
	    if( strcmp( tmp, DIRWORD ) == 0 ) {
		/* It's a directory */
		current->item.dir = 1;
	    } else {
		current->item.dir = 0;
		current->item.mtime = atol( tmp );
	    }
	}
	fclose( fp );
    }
    /* Now read the local files */
    cur_site = the_site; /* So the callback knows what's going on */
    ftw( the_site->local_dir, ftw_callback, 10 );
    return 0;
}

int ftw_callback( const char *file, struct stat *sb, int flag ) {
    char *fname;
    struct ll_files *current;
    /* Get the filename relative to local root - chop the beginning
     * bit off */
    fname = (char *)file+strlen(cur_site->local_dir);

    if( strlen(fname) == 0) return 0;
    if( debug_level==debug_full )
	fprintf( stderr, "Local [%s]: ", fname );

    /* Now process it */
    for( current=cur_site->files; current!=NULL; current=current->next ) {
	if( strcmp( current->item.name, fname ) == 0 )
	    /* Match found! */
	    break;
    }
    
    switch( flag ) {
    case FTW_F:
	if( debug_level==debug_full ) fprintf( stderr, "file - ");
	if( current == NULL ) {
	    /* No match found... insert a new item */
	    current = malloc( sizeof(struct ll_files) );
	    current->next = cur_site->files;
	    cur_site->files = current;
	    current->item.name = strdup( fname ); 
	    current->item.diff = file_new;
	    current->item.mtime = sb->st_mtime;
	    current->item.dir = 0;
	    if( debug_level==debug_full )
		fprintf( stderr, "new: %d\n", (int)sb->st_mtime); 
	} else {
	    /* Match found... compare it */
	    if( current->item.mtime != sb->st_mtime ) {
		current->item.diff = file_changed;
		current->item.mtime = sb->st_mtime;
		if( debug_level==debug_full )
		    fprintf( stderr, "changed (%ld -> %ld)\n", current->item.mtime, sb->st_mtime );
	    } else {
		current->item.diff = file_unchanged;
		if( debug_level==debug_full )
		    printf("unchanged.\n"); 
	    }
	} 
	break;
	/* ftp_put( (char *) file, fname ); */ break;
    case FTW_D:
	if( debug_level==debug_full )
	    printf("directory - "); 	
	if( current == NULL ) {
	    /* New Directory */
	    current = malloc( sizeof(struct ll_files) );
	    current->next = cur_site->files;
	    cur_site->files = current;
	    current->item.name = strdup( fname ); 
	    current->item.dir = 1;
	    current->item.diff = file_new;
	    if( debug_level==debug_full )
		printf("new.\n"); 
	} else {
	    /* Existing directory */
	    current->item.diff = file_unchanged;
	    if( debug_level==debug_full )
		printf("unchanged.\n"); 
	}
#ifdef ALLDEBUG
	printf( "directory\n" );
#endif
	break;
    default:
	printf( "unknown!\n" );
	break;
    }
    /* carry on */
    return 0;
}

/* This writes the remote files list back to the file */
int site_writefiles( struct site_t *the_site ) {
    FILE *fp;
    struct ll_files *current;
    fp = fopen( the_site->infofile, "w" );
    if( fp == NULL ) {
	/* Couldn't open! */
	perror( "fopen" );
	fprintf( stderr, "%s: Could not open site info file for writing: %s\n", progname, the_site->infofile );
	return 1;
    }
    for( current = the_site->files; current!=NULL; current=current->next ) {
	if( current->item.diff != file_deleted ) {
	    if( current->item.dir ) {
		fprintf( fp, "%s\t%s\n", current->item.name, DIRWORD );
	    } else {
		fprintf( fp, "%s\t%ld\n", current->item.name, current->item.mtime );
	    }
	}
    }
    fclose( fp );
    return 0;
}
