/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@weaver.org>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>
#include <ftw.h> 

#define MAXFILES 2000 /* abritrary */

struct site_file_t;
struct site_t;
struct ll_sites;
struct ll_files;

enum file_diff {
    file_unchanged, /* The file has not been changed */
    file_changed, /* The file has been changed, and should be uploaded */
    file_new, /* The file is new locally, and should be uploaded */
    file_deleted,  /* The file has been deleted locally, and should be remove
		   * remotely */
};

struct site_file_t {
    enum file_diff diff;
    int dir; /* This is a directory */
    char *name; /* the filename RELATIVE to root directory */
    time_t mtime; /* the modification time of the file */
};

/* This represents a site, just a load of strings really */
struct site_t {
    char *name; /* symbolic name for site */
    char *server; /* server name */
    char *username; /* username to log in to server */
    char *password; /* password for ditto */
    char *remote_dir; /* root directory of site on server */
    char *local_dir; /* root directory of site locally */
    struct ll_files *files; /* linked list of remote files */
    char *infofile;  /* local storage file in ~/.sitecopy/  */
};

struct ll_sites {
    struct site_t item;
    struct ll_sites *next;
};

struct ll_files {
    struct site_file_t item;
    struct ll_files *next;
};

/* site_readsites will read ~/.sitecopyrc and fill the sites list */
int site_readsites( char *filename, struct ll_sites **sites );

/* This reads off the local files from the disk, and the
 * remote files from the remote files file */
int site_readfiles( struct site_t *the_site );

/* This writes the remote files list back to the file */
int site_writefiles( struct site_t *the_site );

/* Updates the remote files list of the given site.
 * Tricky, this - have to do a load of LIST's, and try and
 * parse the response sensibly */
int site_update( struct site_t *the_site );

/* Try and find a site of given name. Returns non-zero
 * if a site is found, and points the_site to it. Else 
 * returns zero */
int site_find( struct ll_sites **sites, struct site_t **the_site, char *sitename );

int site_verify( struct site_t *any_site );

int site_list( struct site_t *the_site );

/* Syncronizes the remote site with the local copy.
 * This is the core functionality of sitecopy. */
int site_resync( struct site_t *the_site );

int ftw_callback( const char *, struct stat *, int );
